#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

import ctypes
from pathlib2 import Path
import sys, os
import numpy as np

if sys.platform == 'darwin':
    libpath = Path(__file__).parent / 'libccode_c.dylib'
    lib = ctypes.CDLL(str(libpath))
elif os.name == 'nt':
    libpath = Path(__file__).parent.parent.parent.parent / 'bin' / 'ccode_c.dll'
    if not libpath.exists():
        libpath = Path(__file__).parent.parent / 'bin' / 'ccode_c.dll'
    lib = ctypes.WinDLL(str(libpath))
else:
    libpath = Path(__file__).parent / 'libccode_c.so'
    lib = ctypes.CDLL(str(libpath))


asuint = ctypes.c_uint32
asfloat = ctypes.c_float

cfloatp = ctypes.POINTER(ctypes.c_float)
def asfloatp(arr):
    return arr.ctypes.data_as(cfloatp)

cuintp = ctypes.POINTER(ctypes.c_uint32)
def asuintp(arr):
    return arr.ctypes.data_as(cuintp)

lib.newsphere.restype = ctypes.c_uint32
lib.gettouching.restype = ctypes.c_uint32

def drawnewpositions(pos3, ds, zrange):
    lib.drawnewpositions(asfloatp(pos3), asfloatp(ds), asuint(ds.size), asfloat(zrange))

def newsphere(pos3, ds, spheres, zrange, updated):
    return lib.newsphere(asfloatp(pos3), asfloatp(ds), asfloatp(spheres), asuint(ds.size), asuint(spheres.size//5), asfloat(zrange), asuintp(updated))

def gettouching(spheres, i, cutoff):
    touching = np.zeros(spheres.size//5, dtype=np.uint32)
    ntouch =  lib.gettouching(asfloatp(spheres), asuint(spheres.size//5), asuint(i), asfloat(cutoff), asuintp(touching))
    return touching[:ntouch].copy()

def setthreads(nthrds):
    lib.set_threads(asuint(nthrds))

def setseed(seed):
    lib.setseed(asuint(seed))

def genvol(spheres, vol, nx, ny, nz, voxsize, iz, cx=0, cy=0, cz=0, supersampling=1):
    n = np.array([nx,ny,nz],dtype=np.uint32)
    c = np.array([cx,cy,cz],dtype=np.float32)
    lib.genvol(asfloatp(spheres), asuint(spheres.size//5), asfloatp(vol.ravel()), asuintp(n), asfloat(voxsize), asfloatp(c), asuint(supersampling), asuint(iz))

def average2d(vol, supersampling):
    volnew = np.zeros((vol.shape[0]//supersampling, vol.shape[1]//supersampling), dtype=np.float32)
    lib.average2d(asfloatp(vol.ravel()), asfloatp(volnew.ravel()), asuint(volnew.shape[1]), asuint(volnew.shape[0]), asuint(supersampling))
    return volnew


def genparproj(spheres, nx, ny, pixsize, angle, cx=0, cy=0, rotcx=0, rotcy=0):
    proj = np.zeros((ny,nx),dtype=np.float32)
    n = np.array([nx,ny],dtype=np.uint32)
    c = np.array([cx,cy],dtype=np.float32)
    rotc = np.array([rotcx,rotcy],dtype=np.float32)
    lib.genparproj(asfloatp(spheres), asuint(spheres.size//5), asfloatp(proj.ravel()), asuintp(n), asfloat(pixsize), asfloatp(c), asfloat(angle), asfloatp(rotc))
    return proj

def gen3dproj(spheres, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf):
    sph_rot = spheres.copy()
    trot1 = np.array([[1,0,0],[0,np.cos(tilt1), -np.sin(tilt1)],[0,np.sin(tilt1),np.cos(tilt1)]])
    trot2 = np.array([[np.cos(tilt2),0,np.sin(tilt2)],[0,1,0],[-np.sin(tilt2), 0, np.cos(tilt2)]])
    arot = np.array([[np.cos(angle), -np.sin(angle),0],[np.sin(angle),np.cos(angle),0],[0,0,1]])
    rot = np.dot(trot2,np.dot(trot1,arot))
    pos = np.array([spheres[::5],spheres[1::5],spheres[2::5]])
    pos_rot = np.dot(rot,pos)
    trot1 = np.array([[1,0,0],[0,np.cos(-tilt1), -np.sin(-tilt1)],[0,np.sin(-tilt1),np.cos(-tilt1)]])
    trot2 = np.array([[np.cos(-tilt2),0,np.sin(-tilt2)],[0,1,0],[-np.sin(-tilt2), 0, np.cos(-tilt2)]])
    arot = np.array([[np.cos(-angle), -np.sin(-angle),0],[np.sin(-angle),np.cos(-angle),0],[0,0,1]])
    rot = np.dot(arot,np.dot(trot1,trot2)).astype(np.float32)
    sph_rot[::5] = pos_rot[0]
    sph_rot[1::5] = pos_rot[1]
    sph_rot[2::5] = pos_rot[2]
    proj = np.zeros((ny,nx),dtype=np.float32)
    n = np.array([nx,ny],dtype=np.uint32)
    lib.gen3dproj(asfloatp(sph_rot), asuint(spheres.size//5), asfloatp(proj.ravel()), asuintp(n), asfloat(pixsize), asfloat(maxz), asfloatp(rot), asuint(cutout), asfloat(cutoff))
    return proj

def genconeproj(spheres, nx, ny, pixsize, angle, sod, sdd, zoff=0):
    proj = np.zeros((ny,nx),dtype=np.float32)
    n = np.array([nx,ny],dtype=np.uint32)
    ca = np.cos(angle)
    sa = np.sin(angle)
    sph_rot = spheres.copy()
    rotx = ca*sph_rot[::5] - sa*sph_rot[1::5]
    sph_rot[1::5] = sa*sph_rot[::5] + ca*sph_rot[1::5]
    sph_rot[::5] = rotx
    lib.genconeproj(asfloatp(sph_rot), asuint(sph_rot.size//5), asfloatp(proj.ravel()), asuintp(n), asfloat(pixsize), asfloat(zoff), asfloat(sod), asfloat(sdd))
    return proj
    

def applypoisson(proj, flux, factor):
    lib.applypoisson(asfloatp(proj.ravel()), asuint(proj.size), asfloat(flux), asfloat(factor))

# Try to set number of threads to number of physical cores
try:
    import psutil
    ncpu = psutil.cpu_count(logical=False)
    try:
        naff = len(psutil.Process().cpu_affinity())
        if naff < ncpu:
            ncpu = naff
    except AttributeError:
        pass
    setthreads(ncpu)
except ImportError:
    pass

try:
    from numba import cuda, float32, int32
    import math

    def genconeproj_cuda(spheres, nx, ny, pixsize, angle, sod, sdd, zoff=0):
        proj = cuda.to_device(np.zeros((ny,nx),dtype=np.float32))
        bpg0 = (ny + (32 - 1)) // 32
        bpg1 = (nx + (32 - 1)) // 32
        ca = np.cos(angle)
        sa = np.sin(angle)
        sph_rot = spheres.copy()
        rotx = ca*sph_rot[::5] - sa*sph_rot[1::5]
        sph_rot[1::5] = sa*sph_rot[::5] + ca*sph_rot[1::5]
        sph_rot[::5] = rotx
        sph_cuda = cuda.to_device(sph_rot)
        cudaconeproj_impl[(bpg0, bpg1), (32,32)](sph_cuda, proj, np.float32(pixsize), np.float32(sod), np.float32(sdd), np.float32(zoff))
        return proj.copy_to_host()

    @cuda.jit(fastmath=True)
    def cudaconeproj_impl(spheres, proj, pixsize, sod, sdd, zoff):
        iy, ix = cuda.grid(2)
        if iy<proj.shape[0] and ix<proj.shape[1]:
            x = (ix+0.5)*pixsize - proj.shape[1]*pixsize/2
            y = (iy+0.5)*pixsize - proj.shape[0]*pixsize/2
            
            # https://math.stackexchange.com/questions/2613781/line-cylinder-intersection
            x0 = -sod
            k = x/sdd
            l = y/sdd
            df = 1 - (x0*x0 - 1)*k*k
            if df>0:
                t1 = -(math.sqrt(df)+x0)/(k*k+1)
                t2 = (math.sqrt(df)-x0)/(k*k+1)
                dx = t2 - t1
                dy = k*dx
                dz = l*dx
                tmp = math.sqrt(dx*dx+dy*dy+dz*dz)
            else:
                proj[iy,ix]=0
                return

            sz = math.sqrt(1+k*k+l*l)
            tx = 1/sz
            ty = k/sz
            tz = l/sz

            # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
            for j in range(spheres.shape[0]//5):
                sz = spheres[5*j+2]-zoff
                sx = spheres[5*j]
                sy = spheres[5*j+1]
                s2 = spheres[5*j+3]*spheres[5*j+3]
                psz = (x0-sx)*tx - sy*ty - sz*tz
                pdx = x0-sx - psz*tx
                pdy = -sy - psz*ty
                pdz = -sz - psz*tz
                dist = pdx*pdx+pdy*pdy+pdz*pdz
                if dist<s2:
                    tmp -= 2*(1-spheres[5*j+4])*math.sqrt(s2 - dist)
            proj[iy,ix] = tmp
except:
    pass


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

import h5py

class VolumeGeometry(object):
    def __init__(self, nx, ny, nz, voxsize, cx=0, cy=0, cz=0, supersampling=1):
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.voxsize = voxsize
        self.cx = cx
        self.cy = cy
        self.cz = cz
        self.supersampling = supersampling
    
    def to_astra(self, single_slice=False):
        if self.cx != 0 or self.cy != 0 or self.cz != 0:
            raise ValueError('Shifted center not supported yet')
        import astra
        if single_slice:
            return astra.create_vol_geom((self.ny, self.nx))
        else:
            return astra.create_vol_geom((self.nz, self.ny, self.nx))
    
    def to_dict(self):
        dct = {}
        for att in ['nx', 'ny', 'nz', 'voxsize', 'cx', 'cy', 'cz', 'supersampling']:
            dct['volgeom_'+att] = getattr(self,att)
        return dct
    
    @classmethod
    def from_file(cls, fn):
        args = []
        kwargs = {}
        with h5py.File(fn, 'r') as f:
            att = f['volume'].attrs
            for key in ['nx', 'ny', 'nz', 'voxsize']:
                args.append(att['volgeom_'+key])
            for key in ['cx', 'cy', 'cz', 'supersampling']:
                kwargs[key] = att['volgeom_'+key]
        return cls(*args, **kwargs)

class ParallelGeometry(object):

    def __init__(self, nx, ny, angles, pixsize, cx=0, cy=0, rotcx=0, rotcy=0, supersampling=1):
        self.nx = nx
        self.ny = ny
        self.angles = angles
        self.pixsize = pixsize
        self.cx = cx
        self.cy = cy
        self.rotcx = rotcx
        self.rotcy = rotcy
        self.supersampling = supersampling
    
    def to_astra(self, single_slice=False):
        import astra
        if single_slice:
            return astra.create_proj_geom('parallel', 1, self.ny, self.angles)
        else:
            return astra.create_proj_geom('parallel3d', 1, 1, self.ny, self.nx, self.angles)
    
    def to_dict(self):
        dct = {}
        dct['projgeom_type'] = 'parallel'
        for att in ['nx', 'ny', 'angles', 'pixsize', 'cx', 'cy', 'rotcx', 'rotcy', 'supersampling']:
            dct['projgeom_'+att] = getattr(self,att)
        return dct
    
    @classmethod
    def from_file(cls, fn):
        args = []
        kwargs = {}
        with h5py.File(fn, 'r') as f:
            att = f['projs'].attrs
            for key in ['nx', 'ny', 'angles', 'pixsize']:
                if key == 'angles' and 'projgeom_angles' in f.keys():
                    args.append(f['projgeom_angles'][:])
                else:
                    args.append(att['projgeom_'+key])
            for key in ['cx', 'cy', 'rotcx', 'rotcy', 'supersampling']:
                kwargs[key] = att['projgeom_'+key]
        return cls(*args, **kwargs)

class ConeGeometry(object):

    def __init__(self, nx, ny, angles, pixsize, sod=None, sdd=None, odd=None, supersampling=1, zoff=0, usecuda=True):
        self.nx = nx
        self.ny = ny
        self.angles = angles
        self.pixsize = pixsize
        if sod is not None and sdd is not None and odd is not None:
            raise ValueError("Two of SOD, SDD, or ODD have to be given")
        elif sod is not None and sdd is not None:
            self.sod = sod
            self.odd = sdd - sod
        elif sod is not None and odd is not None:
            self.sod = sod
            self.odd = odd
        elif sdd is not None and odd is not None:
            self.sod = sdd - odd
            self.odd = odd
        else:
            raise ValueError("Two of SOD, SDD, or ODD have to be given")
        self.supersampling = supersampling
        self.zoff = zoff
        self.usecuda = usecuda
    
    def to_astra(self):
        import astra
        return astra.create_proj_geom('cone', 1, 1, self.ny, self.nx, self.angles, self.sod/self.pixsize, self.odd/self.pixsize)
    
    def to_dict(self):
        dct = {}
        dct['projgeom_type'] = 'cone'
        for att in ['nx', 'ny', 'angles', 'pixsize', 'sod', 'odd', 'supersampling', 'zoff', 'usecuda']:
            dct['projgeom_'+att] = getattr(self,att)
        return dct
    
    @classmethod
    def from_file(cls, fn, usecuda=True):
        args = []
        kwargs = {}
        with h5py.File(fn, 'r') as f:
            att = f['projs'].attrs
            for key in ['nx', 'ny', 'angles', 'pixsize']:
                if key == 'angles' and 'projgeom_angles' in f.keys():
                    args.append(f['projgeom_angles'][:])
                else:
                    args.append(att['projgeom_'+key])
            for key in ['sod', 'odd', 'supersampling', 'zoff']:
                kwargs[key] = att['projgeom_'+key]
            kwargs['usecuda'] = usecuda
        return cls(*args, **kwargs)


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

from . import generate, infiltrate, project, verticalmovement, expand

import abc
import os.path
import numpy as np

class Phantom(object):
    
    __metaclass__ = abc.ABCMeta

    def __init__(self, filename):
        self.filename = filename
        if not os.path.exists(self.filename):
            raise ValueError("{} does not exist.".format(self.filename))
    
    @abc.abstractmethod
    def generate_projections(self, outfile, geom):
        pass
    
    @abc.abstractmethod
    def generate_volume(self, outfile, geom, time=0):
        pass


class FoamPhantom(Phantom):

    @staticmethod
    def generate(filename, seed, nspheres_per_unit=100000, ntrials_per_unit=1000000, maxsize=0.2, zrange=1.5):
        generate.genphantom(filename, seed, nspheres_per_unit=nspheres_per_unit, ntrials_per_unit=ntrials_per_unit, maxsize=maxsize, zrange=zrange)
    
    def generate_projections(self, outfile, geom):
        generate.gen_dataset(outfile,self.filename, geom)
    
    def generate_volume(self, outfile, geom):
        generate.genvol(outfile, self.filename,geom)
    
    def generate_3d(self, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf):
        return generate.gen3d(self.filename, nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=cutoff)

class MovingFoamPhantom(Phantom):

    @staticmethod
    def generate(filename, phantom_file, seed, zmin, zmax, random_move=0.1, regularization=0.01, npoints=1024):
        verticalmovement.generate_verticalmovement(filename, phantom_file, seed, zmin, zmax, random_move=random_move, regularization=regularization, npoints=npoints)
    
    def generate_projections(self, outfile, geom):
        verticalmovement.gen_dataset_verticalmovement(outfile,self.filename, geom)
    
    def generate_volume(self, outfile, geom, time=0):
        verticalmovement.genvol_verticalmovement(time, outfile, self.filename, geom)
    
    def generate_3d(self, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf, time=0):
        return verticalmovement.gen3d_verticalmovement(time, self.filename, nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=cutoff)

class ExpandingFoamPhantom(Phantom):

    @staticmethod
    def generate(filename, phantom_file, seed, start_size=0.25, random_move=0.1, regularization=0.01, static_after_fraction=0.1, npoints=1024):
        expand.generate_expand(filename, phantom_file, seed, start_size=start_size, random_move=random_move, regularization=regularization, static_after_fraction=0.1, npoints=npoints)
    
    def generate_projections(self, outfile, geom):
        expand.gen_dataset_expand(outfile,self.filename, geom)
    
    def generate_volume(self, outfile, geom, time=0):
        expand.genvol_expand(time, outfile, self.filename, geom)
    
    def generate_3d(self, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf, time=0):
        return expand.gen3d_expand(time, self.filename, nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=cutoff)

class InfiltrationFoamPhantom(Phantom):

    @staticmethod
    def generate(filename, phantom_file, seed, fluid_value, startz=-1.5, rand_width=1, cutoff=1e-5):
        infiltrate.generate_inflitration(filename, phantom_file, seed, fluid_value, startz=startz, rand_width=rand_width, cutoff=cutoff)

    def generate_projections(self, outfile, geom):
        infiltrate.gen_dataset_infiltrate(outfile,self.filename,geom)
    
    def generate_volume(self, outfile, geom, time=0):
        infiltrate.genvol_infiltrate(time, outfile, self.filename,geom)
    
    def generate_3d(self, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf, time=0):
        return infiltrate.gen3d_infiltrate(time, self.filename, nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=cutoff)


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------


from .phantom import FoamPhantom, Phantom, MovingFoamPhantom, InfiltrationFoamPhantom, ExpandingFoamPhantom
from .artifacts import estimate_absorption_factor, apply_poisson_noise
from .geometry import VolumeGeometry, ParallelGeometry, ConeGeometry
from .utils import load_projections, load_volume


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

from .utils import FILE_VERSION
from . import ccode
import h5py
import numpy as np
import scipy.optimize as sco
import tqdm

def apply_poisson_noise(input_file, output_file, seed, flux, absorption_factor):
    ccode.setseed(seed)
    with h5py.File(input_file, 'r') as f:
        projs = f['projs']
        spatt = dict(f['projs'].attrs)
        with h5py.File(output_file, 'w') as fo:
            fo.attrs['FILE_VERSION'] = FILE_VERSION
            dset = fo.create_dataset('projs', projs.shape, dtype='f4')
            for i in tqdm.trange(projs.shape[0]):
                curproj = projs[i][:]
                ccode.applypoisson(curproj,flux,absorption_factor)
                dset[i] = curproj
            att = fo['projs'].attrs
            for key, val in spatt.items():
                att[key] = val
            att['seed'] = seed
            att['flux'] = flux
            att['absorption_factor'] = absorption_factor
            if 'times' in f:
                fo['times'] = f['times'][:]

def estimate_absorption_factor(input_file, average_absorption_ratio):
    with h5py.File(input_file, 'r') as f:
        projs = f['projs'][:]
    projs = projs[projs>0]

    def error_function(fac):
        rat = 1 - np.exp(-projs*fac[0]).mean()
        return (rat - average_absorption_ratio)**2
    
    return sco.fmin(error_function,1)[0]
    


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

import h5py

FILE_VERSION=1.0


def load_volume(fn):  
    with h5py.File(fn,'r') as f:
        vol = f['volume'][:]
    return vol

def load_projections(fn):
    with h5py.File(fn,'r') as f:
        projs = f['projs'][:]
    return projs


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

import h5py
import numpy as np

from . import ccode

def single_par_projection(phantom, nx, ny, pixsize, angle, cx=0, cy=0, rotcx=0, rotcy=0, supersampling=1):
    if isinstance(phantom, str):
        with h5py.File(phantom, 'r') as f:
            spheres = f['spheres'][:]
    else:
        spheres = phantom
    proj = ccode.genparproj(spheres,supersampling*nx,supersampling*ny,pixsize/supersampling,angle,cx,cy,rotcx,rotcy)
    if supersampling>1:
        return ccode.average2d(proj, supersampling)
    return proj

def single_cone_projection(phantom, nx, ny, pixsize, angle, sod, sdd, zoff=0, supersampling=1, usecuda=True):
    if isinstance(phantom, str):
        with h5py.File(phantom, 'r') as f:
            spheres = f['spheres'][:]
    else:
        spheres = phantom
    if usecuda:
        proj = ccode.genconeproj_cuda(spheres, supersampling*nx, supersampling*ny, pixsize/supersampling, angle, sod, sdd, zoff=zoff)
    else:
        proj = ccode.genconeproj(spheres, supersampling*nx, supersampling*ny, pixsize/supersampling, angle, sod, sdd, zoff=zoff)
    if supersampling>1:
        return ccode.average2d(proj, supersampling)
    return proj


#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

import h5py
import numpy as np
import random
import tqdm
from . import ccode,project,generate,geometry
from .utils import FILE_VERSION

def generate_inflitration(outfile, phantom_file, seed, fluid_value, startz=1.5, rand_width=1, cutoff=1e-5):
    random.seed(seed)
    with h5py.File(phantom_file, 'r') as f:
        spheres = f['spheres'][:]
        spatt = dict(f['spheres'].attrs)
    
    nsph = spheres.shape[0]//5

    
    all_touching = []
    for i in range(nsph):
        all_touching.append(ccode.gettouching(spheres,i,cutoff))

    arrival = np.zeros(nsph, dtype=np.float32)
    arrival[:] = np.inf

    to_be_processed = set()
    for i in range(nsph):
        if np.abs(spheres[5*i+2]-startz) < spheres[5*i+3]:
            arrival[i] = 0
            to_be_processed.add(i)
    while(len(to_be_processed)>0):
        proc = to_be_processed.pop()
        for i in all_touching[proc]:
            newtime = arrival[proc] + 1 + (random.random()-0.5)*rand_width
            if newtime<arrival[i]:
                arrival[i] = newtime
                to_be_processed.add(i)
    
    arrival /= arrival[np.isinf(arrival)==False].max()
    
    with h5py.File(outfile, 'w') as f:
        f.attrs['FILE_VERSION'] = FILE_VERSION
        f['spheres'] = spheres
        att = f['spheres'].attrs
        for key, val in spatt.items():
            att[key] = val
        f['arrival'] = arrival
        att = f['arrival'].attrs
        att['seed'] = seed
        att['startz'] = startz
        att['rand_width'] = rand_width
        att['cutoff'] = cutoff
        att['fluid_value'] = fluid_value

def single_projection_infiltrate(time, phantom, geom, angle):
    with h5py.File(phantom, 'r') as f:
        spheres = f['spheres'][:].reshape((-1,5))
        arrival = f['arrival'][:]
        fluid_value = f['arrival'].attrs['fluid_value']
    
    spheres[arrival<=time,-1] = fluid_value
    spheres[arrival>time,-1] = 0

    if type(geom) == geometry.ParallelGeometry:
        return project.single_par_projection(spheres.ravel(), geom.nx, geom.ny, geom.pixsize, angle, geom.cx, geom.cy, geom.rotcx, geom.rotcy, geom.supersampling)
    elif type(geom) == geometry.ConeGeometry:
        return project.single_cone_projection(spheres.ravel(), geom.nx, geom.ny, geom.pixsize, angle, geom.sod, geom.sod + geom.odd, zoff=geom.zoff, supersampling=geom.supersampling, usecuda=geom.usecuda)

def genvol_infiltrate(time, outfile, phantom, geom):
    with h5py.File(phantom, 'r') as f:
        spheres = f['spheres'][:].reshape((-1,5))
        arrival = f['arrival'][:]
        fluid_value = f['arrival'].attrs['fluid_value']
    
    spheres[arrival<=time,-1] = fluid_value
    spheres[arrival>time,-1] = 0
    return generate.genvol(outfile, spheres.ravel(), geom)

def gen3d_infiltrate(time, phantom, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf):
    with h5py.File(phantom, 'r') as f:
        spheres = f['spheres'][:].reshape((-1,5))
        arrival = f['arrival'][:]
        fluid_value = f['arrival'].attrs['fluid_value']
    
    spheres[arrival<=time,-1] = fluid_value
    spheres[arrival>time,-1] = 0
    return ccode.gen3dproj(spheres.ravel(), nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=fluid_value)

def gen_dataset_infiltrate(outfile, phantom, geom):
    angles = geom.angles
    nx = geom.nx
    ny = geom.ny
    times = np.linspace(0, 1, len(angles))
    with h5py.File(outfile, 'w') as f:
        f.attrs['FILE_VERSION'] = FILE_VERSION
        dset = f.create_dataset('projs', (len(angles),ny,nx), dtype='f4')
        for i in tqdm.trange(len(angles)):
            dset[i] = single_projection_infiltrate(times[i],phantom,geom,angles[i])
        att = dset.attrs
        for key, val in geom.to_dict().items():
            att[key] = val
        att['phantom'] = phantom
        att['times'] = times





#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

import numpy as np
import random
import sortedcollections
import tqdm
import h5py
import inspect

from . import ccode, project, geometry
from .utils import FILE_VERSION



def genphantom(outfile, seed, nspheres_per_unit=100000, ntrials_per_unit=1000000, maxsize=0.2, zrange=1.5):
    n = int(ntrials_per_unit*zrange)
    nsph = int(nspheres_per_unit*zrange)
    
    pos3 = np.zeros(n*3,dtype=np.float32)                                                                                                              
    random.seed(seed)
    ccode.setseed(random.randint(0,4294967295))
    
    ds = np.zeros(n,dtype=np.float32)
    
    ccode.drawnewpositions(pos3, ds, zrange)

    if callable(maxsize):
        maxsizes = maxsize(pos3[::3],pos3[1::3], pos3[2::3])
        msk = ds<-maxsizes
        ds[msk] = -maxsizes[msk]

    upd = np.zeros(n, dtype=np.uint32)
    spheres = np.zeros(nsph*5, dtype=np.float32)

    sd = sortedcollections.ValueSortedDict(zip(range(ds.size), ds))

    for i in tqdm.trange(nsph):
        itms = sd.items()
        if callable(maxsize)==False and itms[0][1]<-maxsize:
            allchoices = []
            for itm in itms:
                if itm[1] >= -maxsize:
                    break
                allchoices.append(itm[0])
            ch = random.choice(allchoices)
            spheres[5*i+3] = maxsize
        else:
            allchoices = [itms[0][0],]
            curmax = itms[0][1]
            for itm in range(1,len(itms)):
                if itms[itm][1] == curmax:
                    allchoices.append(itms[itm][0])
                else:
                    break
            ch = random.choice(allchoices)
            spheres[5*i+3] = -sd[ch]
        spheres[5*i] = pos3[3*ch]
        spheres[5*i+1] = pos3[3*ch+1]
        spheres[5*i+2] = pos3[3*ch+2]
        nupd = ccode.newsphere(pos3, ds, spheres[:5*(i+1)], zrange, upd)
        if callable(maxsize):
            maxsizes = maxsize(pos3[3*upd[:nupd]],pos3[3*upd[:nupd]+1],pos3[3*upd[:nupd]+2])
            msk = ds[upd[:nupd]] < -maxsizes
            ds[upd[:nupd][msk]] = -maxsizes[msk]
        for ky in upd[:nupd]:
            sd[ky] = ds[ky]
    
    with h5py.File(outfile,'w') as f:
        f['spheres'] = spheres
        f.attrs['FILE_VERSION'] = FILE_VERSION
        att = f['spheres'].attrs
        att['seed'] = seed
        att['nspheres_per_unit'] = nspheres_per_unit
        att['ntrials_per_unit'] = ntrials_per_unit
        if callable(maxsize):
            att['maxsize'] = inspect.getsource(maxsize)
        else:
            att['maxsize'] = maxsize
        att['zrange'] = zrange


def genvol(outfile, phantom, geom, zoomfactor=1):
    nx = geom.nx
    ny = geom.ny
    nz = geom.nz
    voxsize = geom.voxsize*zoomfactor
    supersampling = geom.supersampling
    if isinstance(phantom, str):
        with h5py.File(phantom, 'r') as f:
            spheres = f['spheres'][:]
    else:
        spheres = phantom
    mi = np.argmin([nx,ny,nz])
    vol = np.zeros((nz, ny, nx), dtype=np.float32)
    for i in tqdm.trange(nz):
        ccode.genvol(spheres, vol, nx, ny, nz, voxsize, i, cx=geom.cx*zoomfactor, cy=geom.cy*zoomfactor, cz=geom.cz*zoomfactor, supersampling=supersampling)

    with h5py.File(outfile, 'w') as f:
        f.attrs['FILE_VERSION'] = FILE_VERSION
        f['volume'] = vol
        att = f['volume'].attrs
        for key, val in geom.to_dict().items():
            att[key] = val
        if isinstance(phantom, str):
            att['phantom'] = phantom

def gen3d(phantom, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf):
    if isinstance(phantom, str):
        with h5py.File(phantom, 'r') as f:
            spheres = f['spheres'][:]
    else:
        spheres = phantom
    return ccode.gen3dproj(spheres, nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=cutoff)

def gen_dataset(outfile, phantom, geom):
    angles = geom.angles
    nx = geom.nx
    ny = geom.ny
    pixsize = geom.pixsize
    supersampling = geom.supersampling
    with h5py.File(outfile, 'w') as f:
        f.attrs['FILE_VERSION'] = FILE_VERSION
        dset = f.create_dataset('projs', (len(angles),ny,nx), dtype='f4')
        for i in tqdm.trange(len(angles)):
            if type(geom) == geometry.ParallelGeometry:
                dset[i] = project.single_par_projection(phantom,nx,ny,pixsize,angles[i],cx=geom.cx,cy=geom.cy,rotcx=geom.rotcx,rotcy=geom.rotcy, supersampling=supersampling)
            elif type(geom) == geometry.ConeGeometry:
                dset[i] = project.single_cone_projection(phantom, nx, ny, pixsize, angles[i], geom.sod, geom.sod + geom.odd, zoff=geom.zoff, supersampling=supersampling, usecuda=geom.usecuda)
        att = dset.attrs
        for key, val in geom.to_dict().items():
            if key == "projgeom_angles":
                f['projgeom_angles'] = val
            else:
                att[key] = val
        att['phantom'] = phantom




#-----------------------------------------------------------------------
#Copyright 2019 Centrum Wiskunde & Informatica, Amsterdam
#
#Author: Daniel M. Pelt
#Contact: D.M.Pelt@cwi.nl
#Website: http://dmpelt.github.io/foam_ct_phantom/
#License: MIT
#
#This file is part of foam_ct_phantom, a Python package for generating
#foam-like phantoms for CT.
#-----------------------------------------------------------------------

from . import project,generate,geometry,ccode
from .utils import FILE_VERSION

import numpy as np
import h5py
import tqdm

def generate_expand(outfile, phantom_file, seed, start_size=0.25, random_move=0.1, regularization=0.01, static_after_fraction=0.1, npoints=1024):
    with h5py.File(phantom_file, 'r') as f:
        spheres = f['spheres'][:]
        spatt = dict(f['spheres'].attrs)

    np.random.seed(seed)
    sizes = np.zeros(npoints, dtype=np.float32)
    v=1
    for i in range(1,npoints): 
        v = v + np.random.normal()*random_move + regularization*(1-v)
        if v<0:
            v = 0
        sizes[i] = sizes[i-1] + v
    sizes /= sizes[-1]
    sizes[:] = start_size + (1-start_size)*sizes
    sizes = np.interp(np.linspace(0,1,npoints), np.linspace(0,1-static_after_fraction,npoints), sizes)
    with h5py.File(outfile, 'w') as f:
        f.attrs['FILE_VERSION'] = FILE_VERSION
        f['spheres'] = spheres
        att = f['spheres'].attrs
        for key, val in spatt.items():
            att[key] = val
        f['sizes'] = sizes
        att = f['sizes'].attrs
        att['seed'] = seed
        att['start_size'] = start_size
        att['random_move'] = random_move
        att['regularization'] = regularization
        att['static_after_fraction'] = static_after_fraction
        att['npoints'] = npoints

def __get_material_volume(spheres):
    minz = (spheres[2::5] - spheres[3::5]).min()
    maxz = (spheres[2::5] + spheres[3::5]).max()
    zdiff = maxz-minz
    cyl_vol = np.pi*zdiff
    sphere_vol = (4*np.pi*spheres[3::5]**3/3).sum()
    return cyl_vol-sphere_vol

def single_projection_expand(time, phantom, geom, angle):
    with h5py.File(phantom, 'r') as f:
        spheres = f['spheres'][:]
        sizes = f['sizes'][:]
    start_vol = __get_material_volume(spheres)
    size_time = np.interp(time, np.linspace(0,1,sizes.size), sizes)
    spheres[3::5] *= size_time
    cur_vol = __get_material_volume(spheres)
    factor = (start_vol/cur_vol)**(1/3)
    if type(geom) == geometry.ParallelGeometry:
        return project.single_par_projection(spheres.ravel(), geom.nx, geom.ny, geom.pixsize/factor, angle, geom.cx/factor, geom.cy/factor, geom.rotcx/factor, geom.rotcy/factor, geom.supersampling)*factor
    elif type(geom) == geometry.ConeGeometry:
        return project.single_cone_projection(spheres.ravel(), geom.nx, geom.ny, geom.pixsize/factor, angle, geom.sod/factor, (geom.sod + geom.odd)/factor, zoff=geom.zoff/factor, supersampling=geom.supersampling, usecuda=geom.usecuda)*factor

def genvol_expand(time, outfile, phantom, geom):
    with h5py.File(phantom, 'r') as f:
        spheres = f['spheres'][:]
        sizes = f['sizes'][:]
    start_vol = __get_material_volume(spheres)
    size_time = np.interp(time, np.linspace(0,1,sizes.size), sizes)
    spheres[3::5] *= size_time
    cur_vol = __get_material_volume(spheres)
    factor = (start_vol/cur_vol)**(1/3)
    generate.genvol(outfile, spheres, geom, zoomfactor=1/factor)

def gen3d_expand(time, phantom, nx, ny, pixsize, angle, tilt1, tilt2, maxz=1.5, cutout=0, cutoff=-np.inf):
    with h5py.File(phantom, 'r') as f:
        spheres = f['spheres'][:]
        sizes = f['sizes'][:]
    start_vol = __get_material_volume(spheres)
    size_time = np.interp(time, np.linspace(0,1,sizes.size), sizes)
    spheres[3::5] *= size_time
    cur_vol = __get_material_volume(spheres)
    factor = (start_vol/cur_vol)**(1/3)
    spheres *= factor
    spheres[4::5] /= factor
    return ccode.gen3dproj(spheres, nx, ny, pixsize, angle, tilt1, tilt2, maxz=maxz, cutout=cutout, cutoff=cutoff)

def gen_dataset_expand(outfile, phantom, geom):
    angles = geom.angles
    ny = geom.ny
    nx = geom.nx
    times = np.linspace(0, 1, len(angles))
    with h5py.File(outfile, 'w') as f:
        f.attrs['FILE_VERSION'] = FILE_VERSION
        dset = f.create_dataset('projs', (len(angles),ny,nx), dtype='f4')
        for i in tqdm.trange(len(angles)):
            dset[i] = single_projection_expand(times[i],phantom,geom,angles[i])
        att = dset.attrs
        for key, val in geom.to_dict().items():
            att[key] = val
        att['phantom'] = phantom
        att['times'] = times